(***********************************************************************)
(*                                                                     *)
(*          Run-time Code Generation Library for Objective Caml        *)
(*                                                                     *)
(*                   Dmitri Lomov (dsl@tepkom.ru)                      *)
(*              St-Petersburg State University, Russia                 *)
(*                                                                     *)
(***********************************************************************)    

(* $Id: eval.ml,v 1.3 2002/04/03 22:00:11 mitya Exp $ *)

open Code
open Bytecode
open Bytegen

let initialized = ref false
let uninit () = 
  if !initialized
  then
    (Bytegen.uninit (); initialized := false)
  else ()
let init () = 
  Bytegen.init (); 
  at_exit uninit;
  initialized := true


(*== common options ==*)

let unsafe_generation = ref false

let var_hash_size = 32

let create_environment () =
  Env.init 
    (fun i -> i mod var_hash_size)
    (fun (i,_) -> i)
    var_hash_size


open GenHelper

let create_gs () = 
{ env = create_environment (); n_args = 0; last_tag = unknown_tag }

let add_var =
  let next_var = ref 0 in
  fun gs -> fun loc ->
    let i = !next_var in
    next_var := i + 1;
    { gs with env = Env.add (i,loc) gs.env }, i

let clear_tag gs = gs.last_tag <- unknown_tag; gs
let set_unboxed_tag gs = gs.last_tag <- unboxed_tag; gs

exception UndefinedVariable of int

(*== Code generation ==*)

(*
 * Handles functions with multiple arguments
 * (Abstr(Abstr(..)) 
 * Returns function body, generation state with params in env
 *)
let rec unwind_abstr cd gs i =
    match cd with
      Abstr f ->
        let gs, id = add_var gs (Param i) in
        unwind_abstr 
          ((Obj.magic f) (Var id)) gs (i+1)
    | _ -> cd, { gs with n_args = i }

exception NotImplemented of int

(*
 * Moves value from a given location to accu.
 *)
let loc_to_accu loc cg =
  match loc with
    Environment(i,_) ->
      (instr (Ienvacc i) cg)
  | Stack(k,_) ->
      let l = getsp cg in
      (instr (Iacc (l-k-1)) cg)
  | Param n ->
      let l = getsp cg in
      (instr (Iacc (l+n)) cg)
(* 
 * Generates code for given tree node.
 * Params:
 *   cd           - code tree
 *   state        - code generation state
 *   expected_tag - "tag" the result expects to have
 *   last_result  - whether this is the last result of a
 *                  function being generated
 * Result of calculation is in accu
 *)
let rec gen cd ((cg,gs) as state) expected_tag last_result  =
  (*
   * Generates return from a function currently generated
   *)
  let gen_return ((cg,gs) as state) =
    if last_result
    then instr (Ireturn ((getsp cg) + gs.n_args)) cg, gs
    else state in
  (*
   * Prepares for apply.
   * Returns closure code, args list (first first), number of args
   *)
  let rec prepare_apply cd args i =
    match cd with
	Apply(f, arg) ->
          prepare_apply (Obj.magic f) (arg::args) (i+1)
      | _ -> 
	  cd, args, i      
  in 
  (*
   * Generates args (first in list is last). Returns (cg,gs)
   *)
  let rec gen_args state = function
      [] -> state
    | arg::t -> 
	let cg, gs = gen (Obj.magic arg) state unknown_tag false in
	gen_args ((instr (Ipush) cg), gs) t
  in
  (*
   * Prepares environment.
   *)
  let prepare_env env base =
    let prepare_env_aux (loc_list,gs_f,i) (id,loc) =
      let tag =
        (match loc with
             Environment(_,tag) -> tag
           | Stack(_,tag) -> tag
           | _ -> unknown_tag) in
      let gs_f = { gs_f with env = Env.add (id, Environment(i,tag)) gs_f.env }
      in
      (loc::loc_list,gs_f,(i+1)) in
    Env.fold env ([],create_gs (),base) prepare_env_aux
  in
  (*
   * Generates code for closure elements
   *)
  let gen_env cg loc_list =
    let push_env_elem cg loc = 
      instr Ipush (loc_to_accu loc cg) in
    List.fold_left push_env_elem cg loc_list
  in
  (*
   * Generates "binary" instruction
   *)
  let gen_bin left i right tag result_tag state =
    let cg, gs = gen right state tag false in
    let cg, gs = gen left ((instr Ipush cg),gs) tag false in
    (instr i cg), (gs.last_tag <- result_tag; gs) 
  and
    (*
     * Generates "binary" instruction, different
     * for boxed and unboxed values
     *)
    gen_bin2 left iboxed iunboxed right result_tag state =
    let (cg, gs) = gen right state unknown_tag false in
    let right_tag = gs.last_tag in
    (* passing right tag to the left argument;
     * quite simple heuristic... *)
    let (cg, gs) = gen left ((instr Ipush cg),gs) right_tag false in
    let left_tag = gs.last_tag in
    ((if (left_tag == unboxed_tag) or (right_tag == unboxed_tag)
      then (instr iunboxed cg) else (instr iboxed cg)),
     (gs.last_tag <- result_tag; gs))    
  and
    gen_if_int_cmp left cmp const then' else' state =
    let (cg,gs) = gen left state unboxed_tag false in
    let (lthen,cg) = newlabel cg in
    let cg = 
      instr (Ibranchintcmp(cmp,const,lthen)) cg in
    let (cg,gs) = gen else' (cg,gs) expected_tag last_result in
    if not last_result
    then
      let (lexit,cg) = newlabel cg in
      let cg = instr (Ilabel lthen) (instr (Ibranch lexit) cg) in
      let (cg,gs) = gen then' (cg,gs) expected_tag false in
      (instr (Ilabel lexit) cg, gs)
    else
      let cg, gs = gen_return (cg,gs) in
      let cg = instr (Ilabel lthen) cg in
      gen_return (gen then' (cg,gs) expected_tag true)      
 in

(*
 * gen body
 *) 
  match cd with 
    Var id ->
      (let loc = 
         match (Env.get gs.env id) with
	   Some (_,loc) -> loc
         | None -> raise (UndefinedVariable id) in
      (match loc with
	Stack(_,tag) -> gs.last_tag <- tag
      | Environment(_,tag) -> gs.last_tag <- tag
      |	_ -> gs.last_tag <- unknown_tag);
      ((loc_to_accu loc cg), gs))

  | Quote o ->
      (let x = Obj.repr o in
      if Obj.is_block x
      then 
        if Obj.size x > 0
        then
          let i,cg = new_external x cg in
          (instr (Igetfield i) (instr (Ienvacc 1) cg)), 
	  (gs.last_tag <- Obj.tag x; gs)
        else
           (instr (Iatom (Obj.tag x)) cg),  (gs.last_tag <- Obj.tag x; gs)
       else
         (instr (Iconst (Obj.obj x)) cg), 
	 (gs.last_tag <- unboxed_tag;gs))
	
  | Apply(_,_) ->
      (* "flatten" the application *)
      let f_code, args0, n_args0 = prepare_apply cd [] 0 in
      (* Analyze the function code for possible inlining *)
      let inlining, args, n_args = 
	Inliner.analyze f_code args0 n_args0 in
      (* Remember the return address, if needed *)
      let prevsp = getsp cg in
      let l,cg = newlabel cg in
      let cg = 
        if n_args > 3
        then instr (Ipush_retaddr l) cg
        else cg in
      (* Generate the args *)
      let state = 
	if n_args > 0
	then gen_args (cg,gs) (List.rev args) 
        else state in
      (* Generate the function closure  or inline the call *)
      let ((cg,gs) as state) = 
	(match inlining with
	  None ->
	    gen (Obj.magic f_code) state Obj.closure_tag false 
	| Some inlining ->
	    Inliner.inline inlining args0 (n_args0 - n_args)  gen state false 
	) in
      (* Generate the application, if needed *)
      if n_args > 0 
      then
	let cg = 
	  if last_result 
	  then
            (instr (Iappterm(n_args, (getsp cg)+gs.n_args)) cg)
	  else
            (instr (Iapply n_args) cg) 
	in
	(setsp prevsp (instr (Ilabel l) cg)), (clear_tag gs)
      else
	(* if we get here, inlining has done all the job *)
	state
	    
  | Abstr _ ->
      let loc_list,gs_f,n = prepare_env gs.env 2 in
      let cg = gen_env cg loc_list in
      let cd, gs_f = unwind_abstr cd gs_f 0 in
      let handle = gen_fun cd gs_f (is_unsafe cg) in
      let cg = instr (Iconst 0) cg in
      let cg = instr Ipush (instr (Iclosure (n-1)) cg) in
      let cg = instr IVreg_closure (instr (Iconst handle) cg) in
      cg,(gs.last_tag <- Obj.closure_tag; gs)
	
  | Fix f ->
      let loc_list, gs_f, n = prepare_env gs.env 3 in
      let cg = gen_env cg loc_list in
      let gs_f,id = (add_var gs_f (Environment(2,Obj.closure_tag))) in
      let cd,gs_f = unwind_abstr (f (Var id)) gs_f 0 in
      let handle = gen_fun cd gs_f (is_unsafe cg) in
      let cg = instr (Iconst 0) (instr Ipush (instr (Iconst 0) cg)) in
      let cg = instr Ipush (instr (Iclosure (n-1)) cg) in
      let cg = instr IVrec_reg_closure (instr (Iconst handle) cg) in
      cg,(gs.last_tag <- Obj.closure_tag; gs)
	
  | Let(arg,f) ->
      let cg,gs = gen (Obj.magic arg) state unknown_tag false in
      let sp = getsp cg in
      let cg = instr Ipush cg 
      and gs,id = add_var gs (Stack(sp,gs.last_tag)) in
      let ((cg,gs) as result) = 
	gen (f (Var id)) (cg,gs) expected_tag last_result in
      if last_result 
      then result
      else (instr (Ipop 1) cg), gs

  | If(Icmp(lft,cmp,(Quote const)),then',else') 
  | If(Compare(lft,cmp, (Quote const)),then',else') 
      when Obj.is_int (Obj.repr const) ->
      gen_if_int_cmp 
	lft (comparison_from_code cmp) 
	(Obj.magic const) then' else' state 

  | If(Icmp((Quote const),cmp,lft),then',else') 
  | If(Compare((Quote const),cmp,lft),then',else') 
      when Obj.is_int (Obj.repr const) ->
      gen_if_int_cmp 
	lft (commute_comparison (comparison_from_code cmp))
	(Obj.magic const) then' else' state 

  | If(Aeq(lft,(Quote const)),then',else') 
  | If(Aeq((Quote const),lft),then',else') ->
	 gen_if_int_cmp lft EQ (Obj.magic const) then' else' state

  | If(Aneq(lft,(Quote const)),then',else') 
  | If(Aneq((Quote const),lft),then',else') ->
	 gen_if_int_cmp lft NEQ (Obj.magic const) then' else' state

	
  | If(cond, then', else') ->
      let cg, gs = gen (Obj.magic cond) state unboxed_tag false in
      let lelse, cg = newlabel cg in
      let cg = instr (Ibranchifnot lelse) cg in
      let cg,gs  = gen then' (cg,gs) expected_tag last_result in 
      if not last_result
      then
	let lexit, cg = newlabel cg in
	let cg = instr (Ibranch lexit) cg in
	let cg = instr (Ilabel lelse) cg in
	let cg, gs = gen else' (cg,gs) expected_tag false in
	let cg = instr (Ilabel lexit) cg in
	cg, gs
      else
	let cg, gs = gen_return (cg,gs) in
	let cg = instr (Ilabel lelse) cg in
	gen_return (gen else' (cg,gs) expected_tag true)

  | Bor(left,right) ->
      let cg,gs = gen (Obj.magic left) state unboxed_tag false in
      let ltrue,cg = newlabel cg in
      let cg = instr (Ibranchif ltrue) cg in
      let cg,gs = gen (Obj.magic right) (cg,gs) unboxed_tag false in
      let cg = (instr (Ilabel ltrue) cg) in
      cg, (set_unboxed_tag gs)

  | Band(left,right) -> 
      let cg,gs = gen (Obj.magic left) state unboxed_tag false in
      let lfalse,cg = newlabel cg in
      let cg = instr (Ibranchifnot lfalse) cg in
      let cg,gs = gen (Obj.magic right) (cg,gs) unboxed_tag false in
      (instr (Ilabel lfalse) cg), (set_unboxed_tag gs)

  | Bnot(left) ->
      let cg,gs = gen (Obj.magic left) state unboxed_tag false in
      instr Iboolnot cg, (set_unboxed_tag gs)

  | For(lower, dir, upper, f) ->
      let cg, gs = gen lower state unboxed_tag false in
      let sp = getsp cg in
      let cg = instr Ipush cg 
      and gs,id = add_var gs (Stack(sp,unboxed_tag)) in
      let body = f (Var id) in
      let cg, gs = gen upper (cg,gs) unboxed_tag false in
      let cg = instr Ipush cg in
      let start_l, cg = newlabel cg in
      let cond_l, cg = newlabel cg in
      let cg = instr (Ibranch cond_l) cg in
      let cg = instr (Ilabel start_l) cg in
      let cg, gs = gen body (cg,gs) unboxed_tag false in
      let delta = (match dir with DTo -> 1 | DDownTo -> -1) in
      let cg = instr (Iassign 1) (instr (Ioffsetint delta) 
				    (instr (Iacc 1) cg)) in
      let cg = instr (Ilabel cond_l) cg in
      let cg = instr (Iacc 2) (instr Ipush (instr (Iacc 0) cg)) in
      let cg = instr (Iintcomp 
			(match dir with DTo -> LE | DDownTo -> GE)) cg in
      let cg = instr (Ibranchif start_l) cg in
      let cg = instr (Ipop 2) cg in
      cg, (set_unboxed_tag gs)      

  | While(cond,body) ->
      let start_l, cg = newlabel cg in
      let cond_l, cg = newlabel cg in
      let cg = instr (Ibranch cond_l) cg in
      let cg = instr (Ilabel start_l) cg in
      let cg,gs = gen (Obj.magic body) (cg,gs) unboxed_tag false in
      let cg = instr (Ilabel cond_l) cg in
      let cg,gs = gen (Obj.magic cond) (cg,gs) unboxed_tag false in
      let cg = instr (Ibranchif start_l) cg in
      cg, (set_unboxed_tag gs)
	
  | Aeq(left,right) ->
      gen_bin left (Iintcomp EQ) right unknown_tag unboxed_tag state

  | Aneq(left,right) ->
      gen_bin left (Iintcomp NEQ) right unknown_tag unboxed_tag state

  | Equal(left,right) ->
      gen_bin2 left (IVequal) (Iintcomp EQ) right unboxed_tag state

  | Compare(left, comparison, right) ->
      let c = comparison_from_code comparison in
      gen_bin2 left (IVcompare c) (Iintcomp c) right unboxed_tag state


  | Ibin(left,op,right) ->
      let i = 
        (match op with
             IADD -> Iaddint
           | ISUB -> Isubint
           | IMUL -> Imulint
           | IDIV -> Idivint
           | IMOD -> Imodint
           | IAND -> Iandint
           | IOR  -> Iorint
	   | IXOR -> Ixorint
	   | ILSL -> Ilslint
	   | ILSR -> Ilslint
	   | IASR -> Iasrint) in
      (match op, left, right with
	   (IADD|ISUB), _, Quote i ->
	     let (cg,gs) = 
	       gen (Obj.magic left) state unboxed_tag false in
	     let offset = 
	       (match op with 
		    IADD -> i 
		  | ISUB -> -i 
		  | _ -> failwith "!!!") in
	     ((instr (Ioffsetint offset) cg), gs)
	 | IADD, Quote i, _ ->
	     let (cg,gs) = 
	       gen (Obj.magic left) state unboxed_tag false in
	     ((instr (Ioffsetint i) cg), gs)
	 | _ ->
	     gen_bin (Obj.magic left) i (Obj.magic right) unboxed_tag
	     unboxed_tag state
      )

  | Iun(op, left) ->
      (match op with
           INEG ->
             let cg,gs = gen (Obj.magic left) state unboxed_tag false in
             (instr Inegint cg), (set_unboxed_tag gs)
	 | INOT ->
             let cg,gs = gen left state unboxed_tag false in
             instr Ixorint 
               (instr (Iconst (lnot 0)) (instr Ipush cg)), 
	       (set_unboxed_tag gs))

  | Icmp(left, cmp, right) ->
      let c = 
        (match cmp with
             Code.EQ -> EQ
           | Code.NEQ -> NEQ
           | Code.LE -> LE
           | Code.GE -> GE
           | Code.LT -> LT
           | Code.GT -> GT) in
      gen_bin left (Iintcomp c) right unboxed_tag unboxed_tag state

  | Flop(left, op, right) ->
      let cg,gs = gen (Obj.magic right) state Obj.double_tag false in
      let cg,gs = gen (Obj.magic left) (instr Ipush cg,gs) 
	Obj.double_tag false in
      (instr (IVflop op) cg, (gs.last_tag <- Obj.double_tag; gs))

  | Float_neg f ->
      let cg,gs = gen (Obj.magic f) state Obj.double_tag false in
      (instr (IVfloat_neg) cg, (gs.last_tag <- Obj.double_tag; gs))

  | Seq(first,second) ->
      gen second 
	(gen (Obj.magic first) state unboxed_tag false) 
	expected_tag
	true

  | ArrayGet(array,index) ->
      let cg,gs = gen index state unboxed_tag false in
      let cg,gs = 
	gen (Obj.magic array) ((instr Ipush cg),gs) unknown_tag false in
      let cg = 
	(if gs.last_tag = Obj.double_array_tag 
	   or expected_tag = Obj.double_tag
 	 then (instr IVgetfloatitem cg)
	 else if gs.last_tag = unknown_tag && expected_tag = unknown_tag
	 then 
	   (instr (IVgetvectitem) cg)
	 else
	   (instr Igetvectitem cg)) in
      cg, clear_tag gs

  | ArraySet(array,index,v) ->
      let cg,gs = gen v state unknown_tag false in
      let item_tag = gs.last_tag in
      let cg,gs = gen index ((instr Ipush cg),gs) unboxed_tag false in
      let cg,gs = 
	gen (Obj.magic array) ((instr Ipush cg),gs) unknown_tag false in
      let cg = 
	if gs.last_tag = Obj.double_array_tag or item_tag = Obj.double_tag
	then instr IVsetfloatitem cg
	else if gs.last_tag = unknown_tag && item_tag = unknown_tag
	then instr IVsetvectitem cg
	else instr Isetvectitem cg in 
      cg, (set_unboxed_tag gs)

  | ArrayLength(array) ->
      let cg,gs = gen (Obj.magic array) state unknown_tag false in
      (instr Ivectlength cg), (set_unboxed_tag gs)

  | StringGet(s,index) ->
      let cg,gs = gen index state unboxed_tag false in
      let cg,gs = 
	gen (Obj.magic s) ((instr Ipush cg),gs) Obj.string_tag false in
      (instr Igetstringchar cg), (set_unboxed_tag gs)

  | StringSet(s,index,v) ->
      let cg,gs = gen (Obj.magic v) state unboxed_tag false in
      let cg,gs = gen index ((instr Ipush cg),gs) unboxed_tag false in
      let cg,gs = 
	gen (Obj.magic s) ((instr Ipush cg),gs) Obj.string_tag false in
      (instr Isetstringchar cg), (set_unboxed_tag gs)

  | Ref(v) ->
      let cg,gs = gen (Obj.magic v) state unknown_tag false in
      instr (Imakeblock(1,ref_tag)) cg, (gs.last_tag <- ref_tag; gs)

  | RefVal(r) ->
      let cg,gs = gen (Obj.magic r) state ref_tag false in
      instr (Igetfield 0) cg, clear_tag gs

  | RefSet(r,v) ->
      let cg,gs = gen v state unknown_tag false in
      let cg,gs = gen (Obj.magic r) ((instr Ipush cg), gs) ref_tag false in
      instr (Isetfield 0) cg, (set_unboxed_tag gs)

  | Tuple l ->
      let rec iter state n =
	if n = 0
	then 
	  gen (Obj.magic (Obj.field l 0)) state unknown_tag false
	else 
	  let cg,gs = gen (Obj.magic(Obj.field l n)) state unknown_tag false in
	  iter (instr Ipush cg, gs) (n-1) in
      
      let len = Obj.size l in
      let cg,gs = iter state (len - 1) in
      (instr (Imakeblock(len, tuple_tag)) cg), 
      (gs.last_tag <- tuple_tag; gs)
      
  | Slice(tuple, i) ->
      let cg,gs = gen (Obj.magic tuple) state tuple_tag false in
      instr (Igetfield i) cg, clear_tag gs

  | _ -> raise (NotImplemented  (Obj.tag (Obj.repr cd)))
      
	 and
    gen_fun cd gs unsafe =
	 let cg = start unsafe in
	 let start_label, cg = newlabel cg in
	 let n_params = gs.n_args in
	 if n_params = 1
	 then
	   let cg = instr (Ilabel start_label) cg in
	   let cg, gs = gen cd (cg,gs) unknown_tag true in
	   let cg = instr (Ireturn n_params) cg in
	   stop start_label cg
	 else
	   let cg = instr (Irestart) cg in
	   let cg = instr (Ilabel start_label) cg in
	   let cg = instr (Igrab (n_params-1)) cg in 
	   let cg, gs = gen cd (cg,gs) unknown_tag true in
	   let cg = instr (Ireturn n_params) cg in
	   stop start_label cg
  

let eval_opts f_code unsafe_gen =
  let gs = create_gs () in
  let cd, gs, is_fix =
    match (Obj.magic f_code) with
      Fix f ->
        let gs, id = add_var gs (Environment(2,Obj.closure_tag)) in
        (f (Var id)), gs, true
    | _ -> (Obj.magic f_code), gs, false in
  let cd, gs = unwind_abstr cd gs 0 in
  let handle = gen_fun cd gs unsafe_gen in
  let closure = Obj.new_block Obj.closure_tag 3 in
  (Obj.magic 
     (if is_fix 
     then (RawBytecode.rec_register_closure handle closure)
     else (RawBytecode.register_closure handle closure)))

let eval f_code = (eval_opts f_code !unsafe_generation)

let set_unsafe v =
  unsafe_generation := v
;;

init ()
