/***********************************************************************/
/*                                                                     */
/*          Run-time Code Generation Library for Objective Caml        */
/*                                                                     */
/*                   Dmitri Lomov (dsl@tepkom.ru)                      */
/*              St-Petersburg State University, Russia                 */
/*                                                                     */
/***********************************************************************/    

/* $Id: rawBytecode.c,v 1.3 2002/04/03 22:00:54 mitya Exp $ */

#include <stdio.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/misc.h>
#include <caml/callback.h>

/*
 * Protective programming
 */

static void InternalError(char * s, int line)
{
  fprintf(stderr, "IE in rawBytecode@%d:%s\n", line, s);
  exit(128);
}

#define ASSERT(cond) if(!(cond)) InternalError(#cond, __LINE__)
#define FALSE 0
#define TRUE  1

/*********************************************************
 * Basic types 
 *********************************************************/
typedef unsigned WORD;
typedef int wosize_t;

struct tagCodeAreaChunk;

typedef struct tagHeapBlock
{
  wosize_t size; /* in words */
  unsigned flags;
  struct tagCodeAreaChunk * chunk;
  union
  {
    union
    {
      struct 
      { 
	/* "Program counter" - where to output next word of code */
	int pc; /* in bytes */
	int wp; /* same in words - not used? */	
      } nf; /* "not finished" blocks */
      struct 
      { 
	/* index of "closure block" - block containig 
	 * external references from this code block.
	 */
	int closure_block; 
	/* offset of entry point */
	int offset; 
      } f; /* "finished" code block */
    } code;
    struct
    {
      /* pointers to the next and previous free blocks in free list */
      struct tagHeapBlock * prev, * next;
    } free;
  } specs;
} HeapBlock;

const int header_size  = sizeof(HeapBlock)/sizeof(WORD) +
  (sizeof(HeapBlock)%sizeof(WORD) == 0 ? 0 : 1);
#define W2B(offset) (offset * sizeof(WORD))

#define PREV(b) ((b)->specs.free.prev)
#define NEXT(b) ((b)->specs.free.next)

#define PC(b) ((b)->specs.code.nf.pc)
#define WP(b) ((b)->specs.code.nf.wp)
#define WPofPC(b) (WP(b) = PC(b)*sizeof(WORD))
#define CHUNK(b) ((b)->chunk)

#define CLOSURE_I(b) ((b)->specs.code.f.closure_block)
#define OFFSET(b) ((b)->specs.code.f.offset)

typedef HeapBlock * pHeapBlock;

/*
 * Flags for heap blocks
 */
enum
{
  /* free block */
  BLOCK_FREE      = 1,     
  /* referred, aka reachable block. Used in blocks GC */
  BLOCK_REFERRED  = 1<<2,
  /* Code block that is not yet finished */
  BLOCK_NOT_READY = 1<<3,
  /* Finished block that is not yet "registered", 
   * i.e. the one for which the closure have not yet been created.
   */
  BLOCK_NOT_REG   = 1<<4
};
#define INIT_FREE_FLAGS (BLOCK_FREE)
#define INIT_CODE_FLAGS (BLOCK_NOT_READY)

/********************************************************
 * Global variables
 ********************************************************/

/*
 * Code area
 */
typedef struct tagCodeAreaChunk
{
  /* chunk memory */
  WORD * code_area;
  /* pointer to the first free block */
  pHeapBlock firstFree;
  /* size of the code area */
  int code_area_size;
  /* next and previous heap chunks */
  struct tagCodeAreaChunk * next, * prev;
} CodeAreaChunk;

CodeAreaChunk * codeArea;
#define HEAP(chunk) ((pHeapBlock)(chunk->code_area))
static int def_code_area_size;
#define NEW_CHUNK_SIZE(size) ((size)+header_size < def_code_area_size ? \
             def_code_area_size : 3*(size)/2+header_size)
/*
 * Closures that reference code blocks
 */
#define LAST_CLOSURE_REF (-1)
static value closure_refs = Val_int(0);
static pHeapBlock * referred_blocks = NULL;
static int n_closure_refs;
static int top_closure_ref;
static int first_closure_ref;

/*
 * "Closure blocks" containing external references
 * from each code block
 */
#define MAX_CODE_SLOTS 32767
static value closure_blocks = Val_int(0);
static pHeapBlock * code_slots;
#define LAST_FREE_SLOT (-1)
static int first_empty_slot = LAST_FREE_SLOT;
static int top_code_slot;
static int n_closure_blocks;

/******************************************************
 * Function prototypes
 ******************************************************/
static void init_code_area_chunk(wosize_t size);
static void remove_code_area_chunk(CodeAreaChunk * chunk);

/*
 * Blocks allocation and freeing inside chunks
 */
static pHeapBlock allocate_block (CodeAreaChunk * chunk,
				  pHeapBlock where,
				  wosize_t size);
static pHeapBlock split_block(CodeAreaChunk * chunk, 
			      pHeapBlock where,
			      wosize_t size);
static pHeapBlock get_block(CodeAreaChunk * chunk,
			    wosize_t size);
static void add_free_block(CodeAreaChunk * chunk, 
			   pHeapBlock p, wosize_t s);
static void make_free(CodeAreaChunk * chunk,
		      pHeapBlock p);

/*
 * Block resizing
 */
static pHeapBlock expand_block(pHeapBlock p, int incr);
static pHeapBlock resize_block(pHeapBlock p, int incr);

/*
 * Blocks freeing
 */
static void code_blocks_gc(void);

/*
 * Code references management
 */
static int realloc_refs(void);

/*
 * Code slots management
 */
static void init_code_slots(void);
static void done_code_slots(void);
static int add_code_slot(pHeapBlock p);
static void remove_code_slot(pHeapBlock p);

/*
 * OCaml primitives
 */
typedef value code_pointer;
value raw_init(value nref, value nslots,
	       value code_area_size);
value raw_uninit(value);

value raw_create(value size);
value raw_put_data(value data, value block);
value raw_put_opcode(value opcode, value block);
value raw_copy_code(code_pointer source, value length, value block);
value raw_get_offset(value block);
value raw_get_size(value block);
value raw_put_data_at(value data, value offset, value block);
value raw_finish(value block, value offset, value n_ext, value ext_list);

value raw_register_closure(value block_code, value closure);
value raw_get_raw_register_closure_index(value);

value raw_gc(value);

value raw_time(value);

/* These functions provide for fast analysis in inlining */
code_pointer raw_get_code_pointer(value); 
value raw_test_instruction(code_pointer, value);
value raw_get_word(code_pointer);
code_pointer raw_advance(code_pointer);
code_pointer raw_advance2(code_pointer);
value raw_diff(code_pointer, code_pointer);

/*
 * Debugging
 */
static void print_free(void);
/********************************************************
 * Implementation
 ********************************************************/
/*
 * Code area chunk initialization
 */
static void init_code_area_chunk(wosize_t size)
{
  CodeAreaChunk * chunk = malloc(sizeof(CodeAreaChunk));
  chunk->code_area_size = size;
  chunk->code_area = malloc(size*sizeof(WORD));
  HEAP(chunk)->flags = INIT_FREE_FLAGS;
  PREV(HEAP(chunk)) = NEXT(HEAP(chunk)) = NULL;
  HEAP(chunk)->size = size;
  chunk->firstFree = HEAP(chunk);

  chunk->prev = NULL;
  chunk->next = codeArea;
  if(codeArea != NULL)
    codeArea->prev = chunk;
  codeArea = chunk;
}

static void remove_code_area_chunk(CodeAreaChunk * chunk)
{
  free(chunk->code_area);
  if(chunk->prev != 0)
    chunk->prev->next = chunk->next;
  else
    codeArea = chunk->next;

  if(chunk->next != 0)
    chunk->next->prev = chunk->prev;

  free(chunk);
}

/*
 * Allocates new code block in given free block without
 * splitting
 */
static pHeapBlock allocate_block(CodeAreaChunk * chunk,
  pHeapBlock where, wosize_t size)
{
  wosize_t s = where->size;

  if(PREV(where) != NULL)
    NEXT(PREV(where)) = NEXT(where);
  else
    chunk->firstFree = NEXT(where);
  if(NEXT(where) != NULL)
    PREV(NEXT(where)) = PREV(where);

  where->size = s;
  where->flags = INIT_CODE_FLAGS;
  PC(where) = header_size;
  WPofPC(where);
  CHUNK(where) = chunk;

  return where;
}

/*
 * Splits given block in two pieces and allocates 
 * code block of given size in first piece 
 */
static pHeapBlock split_block(CodeAreaChunk * chunk,
			      pHeapBlock where, wosize_t size)
{
  wosize_t s = where->size;
  pHeapBlock p = PREV(where), 
             n = NEXT(where);
  pHeapBlock newBlock = (pHeapBlock)(((WORD*)where)+size);

  newBlock->size = s - size;
  newBlock->flags = where->flags;
  PREV(newBlock) = p;
  NEXT(newBlock) = n;

  if(p != NULL)
    NEXT(p) = newBlock;
  else
    chunk->firstFree = newBlock;
  if(n != NULL)
    PREV(n) = newBlock;

  where->size = size;
  where->flags = INIT_CODE_FLAGS;
  PC(where) = header_size;  
  WPofPC(where);
  CHUNK(where) = chunk;
  return where;  
}

#define minSplitSize (header_size + 8)

/*
 * Allocates new code block
 */
static pHeapBlock get_block(CodeAreaChunk * chunk, 
			    wosize_t size)
{
  pHeapBlock next;
  
  size += header_size;
  for(next = chunk->firstFree; next != NULL; next = NEXT(next))
  {
    if(next->size > size)
    {
      if(next->size - size < minSplitSize)
	return allocate_block(chunk, next, size);
      else
	return split_block(chunk, next, size);
    }
  }
  return NULL;
}

/*
 * RTCG system initialization
 */
extern value weak_create(value len);
extern value make_vect(value len, value init);

value raw_init(value nref, value nslots, 
               value a_code_area_size)
{
  CAMLparam3(nref, nslots, a_code_area_size);
  
  closure_refs = 0;
  closure_blocks = 0;

  register_global_root(&closure_refs);
  closure_refs = weak_create(nref);
  referred_blocks = malloc(Int_val(nref)*sizeof(pHeapBlock));
  n_closure_refs = Int_val(nref);

  register_global_root(&closure_blocks);
  closure_blocks = make_vect(nslots, Val_int(0));
  n_closure_blocks = Int_val(nslots);
  
  def_code_area_size = Int_val(a_code_area_size);
  codeArea = NULL;

  init_code_slots();
  top_closure_ref = 1;
  first_closure_ref = LAST_CLOSURE_REF;

  CAMLreturn(Val_unit);
}

static void add_free_block(CodeAreaChunk * chunk, pHeapBlock p, wosize_t s)
{
  p->size = s;
  p->flags = INIT_FREE_FLAGS;
  PREV(p) = NULL;
  NEXT(p) = chunk->firstFree;
  if(chunk->firstFree != NULL)
    PREV(chunk->firstFree) = p;
  chunk->firstFree = p;
}

static void make_free(CodeAreaChunk * chunk, pHeapBlock p)
{
  if(!(p->flags & (BLOCK_FREE|BLOCK_NOT_READY)))
     remove_code_slot(p);
  add_free_block(chunk, p, p->size);
}

static void shrink_block(CodeAreaChunk * chunk, pHeapBlock p, int margin)
{
  int s = p->size;
  ASSERT(margin <= s);
  p->size = margin;
  add_free_block(chunk, (pHeapBlock)(((WORD*)p)+margin), s - margin); 
}

/*
 * Garbage collection
 */
static void code_blocks_gc(void)
{
  int i, all_empty;
  pHeapBlock p;
  CodeAreaChunk * chunk;
   
  /* Mark */
  first_closure_ref = LAST_CLOSURE_REF;
  all_empty = 1;
  for(i = n_closure_refs-1; i > 0; i--)
  {
    value v = Field(closure_refs, i);
    if(v != 0 && Is_block(v))
    {            
      referred_blocks[i]->flags |= BLOCK_REFERRED;
      if(all_empty)
      {
	top_closure_ref = i+1;
	all_empty = 0;
      }
    }
    else if(!all_empty)
    {
      Field(closure_refs, i) = Val_int(first_closure_ref);
      first_closure_ref = i;
    }
  }

  /* Sweep */
  chunk = codeArea; 
  while(chunk != NULL)
  {
    WORD * heap_end = chunk->code_area + chunk->code_area_size;
    pHeapBlock lastFree = NULL; 
    CodeAreaChunk * next;
    int size_free = 0;
    int total_free = 0;

    chunk->firstFree = NULL;
    for(p = (pHeapBlock)chunk->code_area; p < (pHeapBlock)heap_end;)
    {
      if(p->flags & (BLOCK_REFERRED | BLOCK_NOT_READY | BLOCK_NOT_REG))
      {
	if(lastFree != NULL)
	{
	  add_free_block(chunk, lastFree, size_free);
	  total_free += size_free;
	}
	p->flags &= ~BLOCK_REFERRED;
	lastFree = NULL;	
      }
      else 
      {
	if(!(p->flags & BLOCK_FREE))
	  remove_code_slot(p);
	if(lastFree == NULL)
	{
	  lastFree = p;
	  size_free = p->size;
	}
	else size_free += p->size;
      }
      p = (pHeapBlock)((WORD*)p + p->size);
    }
    if(lastFree != NULL)
    {
      add_free_block(chunk, lastFree, size_free);
      total_free += size_free;
    }

    ASSERT(total_free <= chunk->code_area_size);

    next = chunk->next;    
    if(total_free == chunk->code_area_size &&
       (chunk->prev != 0 || chunk->next != 0))
    {
      remove_code_area_chunk(chunk);
    }
    chunk = next;
  }
}

/*
 * Creating new code block
 */
static pHeapBlock get_block_somewhere(int size)
{
  CodeAreaChunk * chunk;

  for(chunk = codeArea; chunk != NULL; chunk = chunk->next)
  {
    pHeapBlock result;
    if((result = get_block(chunk, size)) != NULL)
      return result;
  }
  return NULL;
}

static pHeapBlock create_block(int size)
{
  pHeapBlock result = get_block_somewhere(size);  
  if(result == 0)
  {
    code_blocks_gc();
    result = get_block_somewhere(size);
    if(result == 0)
    {
      init_code_area_chunk(NEW_CHUNK_SIZE(size));
      result = get_block_somewhere(size);
      ASSERT(result != 0);
    }
  }
  return result;
}

value raw_create(value size)
{
  CAMLparam0();
  pHeapBlock result = create_block(Int_val(size));
  CAMLreturn ((value) result);
}

/*
 *  Block expansion
 */

static pHeapBlock resize_block(pHeapBlock p, int incr)
{
  int newSize = p->size + incr;
  pHeapBlock newBlock; 
  CodeAreaChunk * chunk;

  ASSERT(p->flags & BLOCK_NOT_READY);
  newBlock = create_block(newSize);
  memcpy(((WORD*)newBlock)+header_size, ((WORD*)p)+header_size,
	 (PC(p)-header_size)*sizeof(WORD));
  PC(newBlock) = PC(p);
  WP(newBlock) = WP(p);
  
  make_free(CHUNK(p),p);
  return newBlock;
}

/*
 * Uninitializing RTCG system
 */
value raw_uninit(value v)
{
  CodeAreaChunk * chunk;
  remove_global_root(&closure_refs);
  remove_global_root(&closure_blocks);
  free(referred_blocks);
  for(chunk = codeArea; chunk != NULL; chunk = chunk->next)
    remove_code_area_chunk(chunk);
  return Val_unit;
}


static void print_free_chunk(CodeAreaChunk * chunk)
{
  pHeapBlock p = chunk->firstFree;
  fprintf(stderr, "[");
  while(p != NULL)
  {
    fprintf(stderr, "(o=%d,s=%d)", (WORD*)p - chunk->code_area, p->size);
    p = NEXT(p);
  }
  fprintf(stderr, "]\n");
}

static void print_free(void)
{
  CodeAreaChunk * p;
  
  for(p = codeArea; p != 0; p = p->next)
    print_free_chunk(p);
}

static void chunk_walk(CodeAreaChunk * chunk)
{
  pHeapBlock p = (pHeapBlock) chunk->code_area;
  WORD * heap_end = chunk->code_area + chunk->code_area_size;

  fprintf(stderr, "Chunk (at 0x%08X):\n", chunk->code_area);
  for(; p < (pHeapBlock)heap_end; p = (pHeapBlock)(((WORD*)p) + p->size))
  {
    fprintf(stderr,"%p %5d ", (WORD*)p, p->size);
    fprintf(stderr,"%04x ", p->flags );
    if((p->flags & BLOCK_FREE) == 0)
    {
      fprintf(stderr, "in use: ");
      if(p->flags & BLOCK_NOT_READY)
	fprintf(stderr, "pc=%d ", PC(p));
      else
	fprintf(stderr, "ext=%d ", CLOSURE_I(p));
    }
    else
      fprintf(stderr, "free");
    fprintf(stderr, "\n");
  }   
}

static void heap_walk(void)
{
  CodeAreaChunk * p;

  for(p = codeArea; p != 0; p = p->next)
    chunk_walk(p);
}

static void ref_walk(void)
{
  int i;
  fprintf(stderr, "Reftab: first free = %d top = %d\n",
	  first_closure_ref, top_closure_ref);
  for(i = 1; i < n_closure_refs; i++)
  {
    value v;

    v = Field(closure_refs, i);
    if(v == 0) continue;
    fprintf(stderr, "%04d:", i);
    if(Is_block(v))
    {
      fprintf(stderr, "[%x", v);
      if(v != 0)
	fprintf(stderr, ":%p]", (WORD*)referred_blocks[i]);
      else
	fprintf(stderr, "]");
    }
    else
      fprintf(stderr, "(%d) ", Int_val(v));
  }
  fputc('\n', stderr);
}
static void extern_walk(void)
{
  int i;

  fprintf(stderr, "Externs: f=%d t=%d\n", first_empty_slot, top_code_slot);
  for(i = 0; i < n_closure_blocks; i++)
  {
    value v = Field(closure_blocks, i);
    int len, j;

    if(!Is_block(v) || v == 0) continue;
    fprintf(stderr, "%p :", (WORD*)code_slots[i]);
    len = Wosize_val(v);
    for(j = 0; j < len; j++)
      fprintf(stderr, "%x ", Field(v, j));
    fputc('\n', stderr);
  }
}
value raw_test(value v)
{
  heap_walk();
  print_free();
  ref_walk();
  extern_walk();
  return Val_unit;
}

static int try_get_ref(void)
{
  int ref = -1;

  if(top_closure_ref < n_closure_refs)
    ref = top_closure_ref++;
  else if(first_closure_ref > 0)
  {
    ref = first_closure_ref;
    first_closure_ref = 
      Int_val(Field(closure_refs, first_closure_ref));
  }

  return ref;
}

value raw_register_closure(value block_code, value closure)
{
  CAMLparam1(closure);
  int index = Int_val(block_code), ref;
  WORD * p = (WORD*)code_slots[index];

  ref = try_get_ref();

  if(ref < 0)
  {
    code_blocks_gc();
    ref = try_get_ref();
        
    if(ref < 0)
    {
      realloc_refs();
      ref = try_get_ref();
    }
  }

  ASSERT(ref >= 0);

  Store_field(closure_refs, ref, closure);
  Store_field(closure, 0, (value)(p + header_size + OFFSET((pHeapBlock)p)));
  Store_field(closure, 1, Field(closure_blocks, index));
  referred_blocks[ref] = (pHeapBlock)p;
  ((pHeapBlock)p)->flags &= ~BLOCK_NOT_REG;
    
  CAMLreturn(closure);
}

value raw_rec_register_closure(value block_code, value closure)
{
  CAMLparam1(closure);
  int index = Int_val(block_code), ref;
  WORD *  p = (WORD*)code_slots[index];

  ref = try_get_ref();
  if(ref < 0)
  {
    code_blocks_gc();
    ref = try_get_ref();
    if(ref < 0)
    {
      realloc_refs();
      ref = try_get_ref();
    }
  }

  Store_field(closure_refs, ref, closure);
  Store_field(closure, 0, (value)(p + header_size + OFFSET((pHeapBlock)p)));
  Store_field(closure, 1, Field(closure_blocks, index));
  Store_field(closure, 2, closure);
  referred_blocks[ref] = (pHeapBlock)p;
  ((pHeapBlock)p)->flags &= ~BLOCK_NOT_REG;
    
  CAMLreturn(closure);
}

/*
 * Look up for C primitives
 */

typedef value (* c_primitive)();

#if (OCAML_MAJOR >= 3) && (OCAML_MINOR >= 4)
#   define CPRIM builtin_cprim
#else
#   define CPRIM cprim
#endif

static int find_primitive(c_primitive f)
{
  extern c_primitive CPRIM[];
  int index;
    
  for(index = 0; CPRIM[index] != NULL; index++)
    if(CPRIM[index] == f)
      return index;
  return -1;
}
value raw_get_raw_register_closure_index(value v)
{
  static int index = -1;
  if(index < 0)
  {
    index = find_primitive(raw_register_closure);
  }

  return Val_int(index);
}

value raw_get_raw_rec_register_closure_index(value v)
{
  static int index = -1;
  if(index < 0)
     index = find_primitive(raw_rec_register_closure);

  return Val_int(index);
}



#define PRIMITIVE(name) extern value name();                  \
value raw_get_##name##_index(value v)                         \
{                                                             \
    static int index = -1;                                    \
    if(index < 0)                                             \
    {                                                         \
      index = find_primitive(name);                           \
      ASSERT(index >= 0);                                     \
    }                                                         \
    return Val_int(index);                                    \
}

PRIMITIVE(equal)
PRIMITIVE(array_get)
PRIMITIVE(array_set)
PRIMITIVE(array_get_float)
PRIMITIVE(array_set_float)
PRIMITIVE(array_get_addr)
PRIMITIVE(array_set_addr)
PRIMITIVE(string_get)
PRIMITIVE(string_set)

PRIMITIVE(neg_float)
PRIMITIVE(add_float)
PRIMITIVE(sub_float)
PRIMITIVE(mul_float)
PRIMITIVE(div_float)
PRIMITIVE(fmod_float)
PRIMITIVE(power_float)

PRIMITIVE(notequal)
PRIMITIVE(lessthan)
PRIMITIVE(greaterthan)
PRIMITIVE(lessequal)
PRIMITIVE(greaterequal)

value raw_put_data(value data, value block)
{
  pHeapBlock p = (pHeapBlock)block;
  if(PC(p) == p->size)
    p = resize_block(p, p->size);
  /***
  fprintf(stderr, "%d:%d ", PC(p), Int_val(data));
  ***/
  ((WORD*)p)[PC(p)++] = Int_val(data);
  return (value)p;
}

#ifdef THREADED_CODE
char **instr_table;
char *instr_base;
#endif

/***
static
#include "opnames.h"
 ***/

value raw_put_opcode(value data, value block)
{
  opcode_t opcode;
  pHeapBlock p = (pHeapBlock)block;

  if(PC(p) == p->size)
    p = resize_block(p, p->size);
  
  opcode = 
#ifdef THREADED_CODE
    (opcode_t)(instr_table[Int_val(data)] - instr_base);
#else
    Int_val(data);
#endif
  /*** 
  fprintf(stderr, "%d:%s ", PC(p), names_of_instructions[Int_val(data)]);
   ***/

  ((WORD*)p)[PC(p)++] = opcode;
  return (value)p;
}

value raw_copy_code(code_pointer source, value length, value block)
{
  pHeapBlock p = (pHeapBlock)block;
  int len = Int_val(length);

  if(len == 0)
    return (value)p;

  if(PC(p)+len >= p->size)
  {
    int incr = (p->size > len ? p->size : len);
    p = resize_block(p, incr);
  }
  memcpy(((WORD*)p) + PC(p), (opcode_t*)source, len*sizeof(opcode_t));
  /*** 
  fprintf(stderr, "%d:inline[%d] ", PC(p), length);
   ***/
  PC(p) += len;
  return (value)p;
}

value raw_put_data_at(value data, value offset, value block)
{
  ((WORD*)block)[Int_val(offset)+header_size] = Int_val(data);
  return block;
}

value raw_get_offset(value block)
{
  int pc = PC((pHeapBlock)block) - header_size;
  return Val_int(pc);
}

value raw_get_size(value block)
{
  wosize_t s = ((pHeapBlock)block)->size - header_size;
  return Val_int(s);
}

value raw_move_to_offset(value block, value off)
{
  int o = Int_val(off) + header_size;
  PC((pHeapBlock)block) = o;
  WPofPC((pHeapBlock)block);
  return Val_unit;
}

#ifdef DEBUG
extern void heap_check(void);
value raw_heapcheck(value v)
{
  heap_check();
  return Val_unit;
}
#endif

value raw_finish(value block, value offset, value n_ext, value ext_list)
{
  CAMLparam1(ext_list);
  CAMLlocal2(externals, next);
  int index;
  pHeapBlock p = (pHeapBlock)block;

  if(p->size - PC(p) > minSplitSize)
    shrink_block(CHUNK(p),p,PC(p));

  index = add_code_slot(p);
  ASSERT(index >= 0);  

  if(n_ext == Val_int(0))
    externals = Atom(0);
  else
  {
    int i; 
    externals = alloc_tuple(Int_val(n_ext));
    next = ext_list;
    for(i = Int_val(n_ext) - 1; i >= 0; i--)
    {
      ASSERT(Is_block(next));
      Store_field(externals, i, Field(next, 0));
      next = Field(next, 1);
    }
  }
  Store_field(closure_blocks, index, externals);
  CLOSURE_I(p) = index;
  OFFSET(p) = Int_val(offset);
  p->flags &= ~BLOCK_NOT_READY;
  p->flags |= BLOCK_NOT_REG;
  CAMLreturn(Val_int(index));
}

value raw_gc(value v)
{
  code_blocks_gc();
  return Val_unit;
}

/*
 * 
 */ 

code_pointer raw_get_code_pointer(value closure)
{
  CAMLparam1(closure);
  CAMLreturn(Field(closure,0));
}


value raw_test_instruction(code_pointer pc, value instr)
{
  opcode_t opcode = Int_val(instr);
#ifdef THREADED_CODE
  return 
    ((instr_base + *((opcode_t*)pc)) == instr_table[opcode] ? 
     Val_true : Val_false);
#else
  return 
    (*((opcode_t*)pc) == opcode ? Val_true :  Val_false);
#endif
}

value raw_get_word(code_pointer pc)
{
  return Val_int(*((opcode_t*)pc));
}

code_pointer raw_advance(code_pointer pc)
{
  return (value)(((opcode_t*)pc) + 1);
}

code_pointer raw_advance2(code_pointer pc)
{
  return (value)(((opcode_t*)pc) + 2);
}

value raw_diff(code_pointer pc1, code_pointer pc2)
{
  int diff = ((opcode_t*)pc2) - ((opcode_t*)pc1);
  return Val_int(diff);
}


/*
 * Code references management
 */

int realloc_refs(void)
{
  CAMLparam0();
  CAMLlocal1(new_refs);
  pHeapBlock * new_referred;
  int new_size = 2*n_closure_refs;
  int i;
  
  new_referred = realloc(referred_blocks, new_size*sizeof(pHeapBlock));
  
  ASSERT(new_referred != 0);
  referred_blocks = new_referred;
  
  
  new_refs = weak_create(Val_int(new_size));
  
  for(i = 1; i < n_closure_refs; i++)
  {
    modify(&Field(new_refs,i), Field(closure_refs,i));
  }

  closure_refs = new_refs;
  n_closure_refs = new_size;

  CAMLreturn(Val_int(0));
}

/*
 * Code slots management
 */
void init_code_slots(void)
{
  code_slots = malloc(sizeof(pHeapBlock)*n_closure_blocks);
  first_empty_slot = LAST_FREE_SLOT;
  top_code_slot = 0;
}
void done_code_slots(void)
{
  free(code_slots);
}

int realloc_code_slots(void)
{
  CAMLparam0();
  CAMLlocal1(new_closures);
  pHeapBlock * new_slots;
  int new_size = n_closure_blocks + n_closure_blocks/2;
  int i;

  if(new_size > MAX_CODE_SLOTS)
    return FALSE;

  new_slots = realloc(code_slots, new_size * sizeof(pHeapBlock));

  if(new_slots == 0)
    return FALSE;
  code_slots = new_slots;
  
  new_closures = alloc_shr(new_size, Tag_val(closure_blocks));
  for(i = 0; i < n_closure_blocks; i++)
    initialize(&Field(new_closures, i), Field(closure_blocks,i));
  for(i = n_closure_blocks; i < new_size; i++)
    initialize(&Field(new_closures, i), 0);
  closure_blocks = new_closures;
  n_closure_blocks = new_size;
  CAMLreturn(TRUE);
}

static int try_alloc_code_slot(void)
{
  int index = LAST_FREE_SLOT;
  if(top_code_slot < n_closure_blocks)
      index = top_code_slot++;
  else if(first_empty_slot >= 0)
  {
    index = first_empty_slot;
    first_empty_slot = (int)code_slots[first_empty_slot];
  }
  return index;
}

static int add_code_slot(pHeapBlock p)
{
  int index = LAST_FREE_SLOT;

  ASSERT((p->flags & BLOCK_FREE) == 0);

  do
  {
    if((index = try_alloc_code_slot()) < 0)
    {
      code_blocks_gc();
      index = try_alloc_code_slot();	
    }
  }
  while(index < 0 && realloc_code_slots());
  
  if(index < 0)
    return LAST_FREE_SLOT;
    
  code_slots[index] = p;
  CLOSURE_I(p) = index;
  return index;     
}

static void remove_code_slot(pHeapBlock p)
{
  int index = CLOSURE_I(p);
  
  Store_field(closure_blocks, index, Val_int(0));
  code_slots[index] = (pHeapBlock)first_empty_slot;
  first_empty_slot = index;
}

/*
 *  Revsion log:
 *  $Log: rawBytecode.c,v $
 *  Revision 1.3  2002/04/03 22:00:54  mitya
 *  Bug in realloc_ref fixed
 *
 *  Revision 1.2  2002/03/18 09:14:28  mitya
 *  Comments added
 *
 *  Revision 1.1.1.1  2002/02/28 11:11:01  mitya
 *  imported
 *
 *  Revision 1.4  2000/12/27 12:47:26  dsl
 *  Improved heap-growing algorithms
 *
 *
 */














