(***********************************************************************)
(*                                                                     *)
(*          Run-time Code Generation Library for Objective Caml        *)
(*                                                                     *)
(*                   Dmitri Lomov (dsl@tepkom.ru)                      *)
(*              St-Petersburg State University, Russia                 *)
(*                                                                     *)
(***********************************************************************)    

(* $Id:*)

(* rtcg.ml *)
open Code

type 'a code = 'a Code.code

(************************************************
 *  Code creation functions
 ************************************************)
let abstr f = Abstr (Obj.magic f)
let apply f x = Apply ((Obj.magic f), (Obj.magic x))
let quote x = Quote x
let fix f = Fix f 
let let' u f = Let ((Obj.magic u), (Obj.magic f))

type direction = DTo | DDownTo
type comparison = LT | GT | LE | GE | EQ | NEQ
let get_cmp = function 
    EQ -> Code.EQ
  | NEQ -> Code.NEQ
  | LT -> Code.LT
  | GT -> Code.GT
  | LE -> Code.LE
  | GE -> Code.GE


let if' cond then' else' =
  If((Obj.magic cond), then',else')
let bor left right =
  Bor(left,right)
let band left right =
  Band(left,right)
let bnot left = 
  Bnot left

let unsafe_tuple l =
  (Obj.repr (Tuple  l))
let unsafe_slice l i =
  (Obj.repr (Slice((Obj.magic l),i)))

let tuple2 a b =
  (Obj.magic (unsafe_tuple (Obj.magic (a,b))))
let tuple3 a b c =
  (Obj.magic (unsafe_tuple (Obj.magic (a,b,c))))

let slice20 t =
  (Obj.magic (unsafe_slice (Obj.magic t) 0))

let slice21 t =
  (Obj.magic (unsafe_slice (Obj.magic t) 1))

let slice30 t =
  (Obj.magic (unsafe_slice (Obj.magic t) 0))

let slice31 t =
  (Obj.magic (unsafe_slice (Obj.magic t) 1))

let slice32 t =
  (Obj.magic (unsafe_slice (Obj.magic t) 2))

let for' lower dir upper body =
  For(
    lower, 
    (match dir with 
      DTo -> Code.DTo
    | DDownTo -> Code.DDownTo),
    upper, body)

let while' cond body = While(cond,body)

let seq first second = Seq(first,second)
let lose c = (Obj.magic c)

let array_set a index v =
  (Obj.magic (ArraySet((Obj.magic a),index,v)))
let array_get a index =
  ArrayGet(a,index)
let array_length a =
  ArrayLength (Obj.magic a)

let string_set s index c =
  (Obj.magic (StringSet(s,index,c)))
let string_get s index =
  StringGet(s,index)

let ref_create v =
  Ref (Obj.magic v)
let ref_value r =
  RefVal r
let ref_set r v =
  RefSet((Obj.magic r),(Obj.magic v))  

let eq left right =
  Aeq((Obj.magic left),(Obj.magic right)) 
let neq left right =
  Aneq((Obj.magic left),(Obj.magic right))
let equal left right =
  Equal((Obj.magic left), (Obj.magic right))
let compare left comparison right =
  Compare((Obj.magic left), (get_cmp comparison), (Obj.magic right))


type int_bin_op = 
  IOpSUB | IOpADD | IOpMUL | IOpDIV | IOpMOD |
  IOpAND | IOpOR  | IOpXOR |
  IOpLSL | IOpLSR | IOpASR 

let get_int_bin_op = function 
  IOpSUB -> ISUB
| IOpADD -> IADD
| IOpMUL -> IMUL
| IOpDIV -> IDIV
| IOpMOD -> IMOD
| IOpAND -> IAND
| IOpOR  -> IOR
| IOpXOR -> IXOR
| IOpLSL -> ILSL
| IOpLSR -> ILSR
| IOpASR -> IASR

type int_un_op = IUnNEG | IUnNOT
let  get_int_un_op = function
  IUnNEG -> INEG
| IUnNOT -> INOT

let int_bin left op right =
  Ibin(left, (get_int_bin_op op), right)

let int_un op left =
  Iun((get_int_un_op op), left)

let int_cmp left cmp right =
  Icmp(left, (get_cmp cmp), right)


type flop = 
    FLOP_ADD | FLOP_SUB | FLOP_MUL | FLOP_DIV | FLOP_MOD | FLOP_POWER

let flop_rtcg2code = function
    FLOP_ADD -> FO_ADD
  | FLOP_SUB -> FO_SUB
  | FLOP_MUL -> FO_MUL
  | FLOP_DIV -> FO_DIV
  | FLOP_MOD -> FO_MOD
  | FLOP_POWER -> FO_POWER

let float_bin left f right =
  Flop(left, (flop_rtcg2code f), right)

let float_neg arg = Float_neg arg
